// -------------------------------------------------------------
// MYENGINE3D v2008.2.20 by Georgy Moshkin
// licence: public domain
// -------------------------------------------------------------
unit mydepthcollision;

interface

uses mybounds,
     math,
     OpenGL;

const DebugOutput=false;

const ZBuf_w = 32;
      ZBuf_h = 32;

const ZAreaElementDimension_x = 50;
      ZAreaElementDimension_z = 50;

const my_NEAR=1;
      my_FAR=1000;

const EPSILON=0.05;
      EPSILON2=1;  

type TZbuffer=packed array[0..ZBuf_w-1,0..ZBuf_h-1] of glFloat;

type TZLayerElement=record
                     zDATA        : TZbuffer;
                     captureHeight: single;
                     parent_i     : integer;
                     parent_j     : integer;
                     parent_k     : integer;
                    end;

type TZLayerElements=array of TZLayerElement;

type TLinkToZLayerElement=record
                           child_index   : integer;
                           captureHeight : single;
                          end;

type TZArea=array of
                  array of
                        array of TLinkToZLayerElement;


type TOldVersionCircleElement=record
                               Radius:single;
                               AntiVectorX:single;
                               AntiVectorZ:single;
                               Valid:boolean;
                              end;

const OldVersionKRUG_w = 8;
      OldVersionKRUG_h = 8;


type TOldVersionCollisionCircle=array[0..OldVersionKRUG_w-1,0..OldVersionKRUG_h-1] of TOldVersionCircleElement;
var TestCircle:TOldVersionCollisionCircle;


var ZBuffer:TZbuffer;
    ZArea:TZArea;
    ZLayerElements:TZLayerElements;

var min_Y, max_Y:single;

var depthcollisionFirstRun:boolean=true;

var z_i, z_j, z_k:integer;


var ZAreaDimension_x: single;
    ZAreaDimension_z: single;

var ZLayer_w: integer;
    ZLayer_h: integer;

var ZBufElementDimension_x:single;
    ZBufElementDimension_z:single;

var ZArea_x:integer;
    ZArea_y:integer;
    ZArea_z:integer;

var ZAreaOffset:TBoundOffset;

procedure FixObjectBounds(var bound3d:TBounds);
procedure DepthCollisionUseBounds(bound3d:TBounds);
procedure BeforeRender;
function AfterRender:boolean;
procedure SetOrthoMode;
procedure SetOrthoCamera(Area_i,Area_j,Area_k:integer);
procedure GetCurrentZBuffer;
procedure StoreBufferData(Area_i,Area_j,Area_k:integer);
procedure SaveData(fname:string;bound3d:TBounds);
procedure LoadData(fname:string;var bound3d:TBounds);
procedure ShowCollisionDataOpengl(ZoneI,ZoneJ,ZoneK:integer; ZoneU,ZoneV:integer; player_x,player_y,player_z:single);
function isequal(zbuf1,zbuf2:TZbuffer):boolean;
function getDIFF(zbuf1,zbuf2:TZbuffer):single;
function DoCollisionEngine(ColCircle:TOldVersionCollisionCircle; CollisionElapsed:single;var Player_x,Player_y,Player_z,old_x,old_y,old_z:single):boolean;
function FindGoodLayerY(iii,kkk,uuu,vvv:integer; Player_y:single):integer;
procedure MakeCollisionCircle(var ColCircle:TOldVersionCollisionCircle);
implementation

procedure MakeCollisionCircle(var ColCircle:TOldVersionCollisionCircle);
var i,j:integer;
    tmpx,tmpz,radius:single;
begin

 for i:=0 to OldVersionKRUG_W-1 do
  for j:=0 to OldVersionKRUG_H-1 do
   begin
    tmpx:=i-(OldVersionKRUG_W-1)/2;
    tmpz:=j-(OldVersionKRUG_H-1)/2;

    radius:=sqrt(tmpx*tmpx+tmpz*tmpz);

    ColCircle[i,j].Valid:=false;

    if radius = 0 then continue;
    if radius > OldVersionKRUG_W div 2{?} then continue;

    ColCircle[i,j].Radius:=radius;

    ColCircle[i,j].Valid:=true;

    ColCircle[i,j].AntiVectorX:=-(tmpx/radius)*( (OldVersionKRUG_W)/2-radius );
    ColCircle[i,j].AntiVectorZ:=-(tmpz/radius)*( (OldVersionKRUG_H)/2-radius );

   end;



end;

function CheckIJKUV(iii,jjj,kkk,uuu,vvv:integer):boolean;
begin

if (iii>-1)and(iii<ZArea_x)and
   (jjj>-1)and(jjj<ZArea_y)and
   (kkk>-1)and(kkk<ZArea_z)and
   (uuu>-1)and(uuu<ZBuf_w)and
   (vvv>-1)and(vvv<ZBuf_h)then result:=true else result:=false;
//result:=true;
end;


function GetElementHeight(iii,jjj,kkk,uuu,vvv:integer):single;
begin
if CheckIJKUV(iii,jjj,kkk,uuu,vvv) then
result:=(ZArea[iii,jjj,kkk].captureHeight - ZLayerElements[ZArea[iii,jjj,kkk].child_index].zDATA[uuu,vvv]*(MY_FAR-MY_NEAR)-{1}MY_NEAR)
end;

function FindGoodLayerY(iii,kkk,uuu,vvv:integer; Player_y:single):integer;
var j:integer;
var good:integer;
    delta:single;
begin

if not CheckIJKUV(iii,0,kkk,uuu,vvv) then exit;
 good:=-12345;
 delta:=1000000;

 for j:=0 to ZArea_y-1 do
   if (Player_y<ZArea[iii,j,kkk].captureHeight) then
    if ({abs}(ZArea[iii,j,kkk].captureHeight - Player_y))<delta then
       begin
        good:=j;
        delta:={abs}(ZArea[iii,j,kkk].captureHeight - Player_y);
       end;


 // under-layer
 if Player_Y < getElementHeight(iii,good,kkk,uuu,vvv) then
 good:=good-1;

// if good=-12345 then halt; //?
 if good=-12345 then good:=0;
 result:= good;

end;

procedure GetIJKUVfromXYZ(var ZoneI,ZoneJ,ZoneK,ZoneU,ZoneV:integer;player_x,player_y,player_z:single);
var PX,PY,PZ:single;
var tmp1,tmp2:single;
begin

PX:=Player_x {- ZAreaOffset.x*0};
PZ:=Player_z {- ZAreaOffset.z*0};

PY:=Player_y;


tmp1:=PX/ZAreaElementDimension_x - (- 0.5*ZAreaDimension_x + 0.5*ZBufElementDimension_x + ZAreaOffset.x)/ZAreaElementDimension_x;
tmp2:=PZ/ZAreaElementDimension_z - (- 0.5*ZAreaDimension_z + 0.5*ZBufElementDimension_z + ZAreaOffset.z)/ZAreaElementDimension_z;


//tmp1:=( Player_X -(- ZAreaElementDimension_x/2 + 0.5*ZBufElementDimension_x + ZAreaOffset.x) )/ZAreaElementDimension_x;
//tmp2:=( Player_Z -(- ZAreaElementDimension_z/2 + 0.5*ZBufElementDimension_z + ZAreaOffset.z) )/ZAreaElementDimension_z;



ZoneI:=round(tmp1);
ZoneK:=round(tmp2);

ZoneU:=ZBuf_w div 2+trunc{?}((tmp1-ZoneI)*ZBuf_w);
ZoneV:=ZBuf_h div 2+trunc{?}((tmp2-ZoneK)*ZBuf_h); //runding bug

//ZoneU:=round(Frac(tmp1)*ZBuf_w);
//ZoneV:=round(Frac(tmp2)*ZBuf_h);
//ZoneU:=round((tmp1-ZoneI)*ZBuf_w);
//ZoneV:=round((tmp2-ZoneK)*ZBuf_h);

ZoneJ:=FindGoodLayerY(ZoneI,ZoneK,ZoneU,ZoneV,
                      PY);


end;

procedure GetXYZfromIJK(var tmpx,tmpy,tmpz:single;Area_i,Area_j,Area_k:integer);
begin

    if not CheckIJKUV(Area_i,Area_j,Area_k,0,0)then exit;

     tmpY:=ZArea[Area_i,Area_j,Area_k].captureHeight;

//     tmpX:=Area_i*ZAreaElementDimension_x - ZAreaElementDimension_x/2 + 0.5*ZBufElementDimension_x + ZAreaOffset.x;
//     tmpZ:=Area_k*ZAreaElementDimension_z - ZAreaElementDimension_z/2 + 0.5*ZBufElementDimension_z + ZAreaOffset.z;
     tmpX:=Area_i*ZAreaElementDimension_x - 0.5*ZAreaDimension_x + 0.5*ZBufElementDimension_x + ZAreaOffset.x;
     tmpZ:=Area_k*ZAreaElementDimension_z - 0.5*ZAreaDimension_z + 0.5*ZBufElementDimension_z + ZAreaOffset.z;

end;

procedure GetXYZfromIJKUV(var x,y,z:single;Area_i,Area_j,Area_k,u,v:integer);
var tmpx,tmpy,tmpz:single;
begin

    if not CheckIJKUV(Area_i,Area_j,Area_k,u,v)then exit;

    GetXYZfromIJK(tmpx,tmpy,tmpz,
                  Area_i,Area_j,Area_k);

     x:=tmpx-ZAreaElementDimension_x/2+u*ZBufElementDimension_x;
     y:=GetElementHeight(Area_i,Area_j,Area_k,u,v);
     z:=tmpz-ZAreaElementDimension_z/2+v*ZBufElementDimension_z;
end;

function DoCollisionEngine(ColCircle:TOldVersionCollisionCircle; CollisionElapsed:single;var Player_x,Player_y,Player_z,old_x,old_y,old_z:single):boolean;
var cld:boolean;
var ZoneI,ZoneJ,ZoneK:integer;
var ZoneU,ZoneV:integer;
var i,j:integer;
var tmp1,tmp2:single;
var ZoneCenterJ:integer;
var OldCenterDepth,CircleElementDepth:single;
var antiHang:integer;


    BestIndexI, BestIndexJ :integer;
    BestIndexRadius:single;

begin

antiHang:=0;

GetIJKUVfromXYZ(ZoneI,ZoneCenterJ,ZoneK,
                ZoneU,ZoneV,
                old_x,old_y,old_z);

OldCenterDepth:=GetElementHeight(ZoneI,ZoneCenterJ,ZoneK,
                                 ZoneU,ZoneV);


glDisable(GL_DEPTH_TEST);

if DebugOutput then
 ShowCollisionDataOpengl(ZoneI,ZoneCenterJ,ZoneK,
                         ZoneU,ZoneV,
                         player_x,player_y,player_z);


repeat
cld:=false;
BestIndexRadius:=1000000;

for i:=0 to OldVersionKRUG_W-1 do
 for j:=0 to OldVersionKRUG_H-1 do
  if TestCircle[i,j].Valid then
   begin
    tmp1:=(Player_x) - OldVersionKRUG_W*ZBufElementDimension_x/2 + 0.5*ZBufElementDimension_x+i*ZBufElementDimension_x;
    tmp2:=(Player_z) - OldVersionKRUG_H*ZBufElementDimension_z/2 + 0.5*ZBufElementDimension_z+j*ZBufElementDimension_z;

    GetIJKUVfromXYZ(ZoneI,ZoneJ,ZoneK,
                    ZoneU,ZoneV,
                    tmp1,Player_y,tmp2);

    CircleElementDepth:=GetElementHeight(ZoneI,ZoneJ,ZoneK,ZoneU,ZoneV);

     if abs(CircleElementDepth-OldCenterDepth)>3 then
      begin
       if CircleElementDepth-OldCenterDepth>0 then glColor3f(1,0,0);
       if CircleElementDepth-OldCenterDepth<0 then glColor3f(0,0,1);

       if TestCircle[i,j].Radius<BestIndexRadius then
        begin
         BestIndexRadius:=TestCircle[i,j].Radius;
         BestIndexI:=i;
         BestIndexJ:=j;
        end;

       Player_x:=Player_x+TestCircle[i,j].AntiVectorX/10;
       Player_z:=Player_z+TestCircle[i,j].AntiVectorZ/10;
       cld:=true;

      end else glColor3f(1,1,1);


    if DebugOutput then
     begin
      glPointSize(3);

      glBegin(GL_POINTS);
      glVertex3f(tmp1,Player_y,tmp2);
      glVertex3f(tmp1,CircleElementDepth,tmp2);
      glEnd;

      glPointSize(1);

      glBegin(GL_LINES);
      glVertex3f(tmp1,CircleElementDepth,tmp2);
      glVertex3f(tmp1+TestCircle[i,j].AntiVectorX,CircleElementDepth,tmp2+TestCircle[i,j].AntiVectorZ);
      glEnd;
     end;

   end;

    i:=BestIndexI;
    j:=BestIndexJ;
    if BestIndexRadius<1000000 then
     begin
//       Player_x:=Player_x+TestCircle[i,j].AntiVectorX/100;
//       Player_z:=Player_z+TestCircle[i,j].AntiVectorZ/100;
     end;


 inc(antiHang);
 if antiHang>0 then break;
 until cld=false;

//player_y:=player_y-0.1*sign(player_y-OldCenterDepth-0.5);
//player_y:=OldCenterDepth+5;

glEnable(GL_DEPTH_TEST);

//if not cld then
begin

GetIJKUVfromXYZ(ZoneI,ZoneCenterJ,ZoneK,
                ZoneU,ZoneV,
                player_x,player_y,player_z);

player_y:=GetElementHeight(ZoneI,ZoneCenterJ,ZoneK,
                                 ZoneU,ZoneV)+4;

//player_x:=old_x+(player_x-old_x)*0.5;
//player_y:=old_y+(player_y-old_y)*0.1;
//player_z:=old_z+(player_z-old_z)*0.5;

(*
if (old_x-player_x)*(old_x-player_x)+
   (old_y-player_y)*(old_y-player_y)+
   (old_z-player_z)*(old_z-player_z)< ZAreaElementDimension_x/100 then
   begin
player_x:=old_x;
player_y:=old_y;
player_z:=old_z;
end;
*)

old_X:=player_x;
old_Y:=player_y;
old_Z:=player_z;

end;

result:=cld;
end;


procedure FixObjectBounds(var bound3d:TBounds);
begin
bound3d.minx:=bound3d.minx-ZAreaElementDimension_x*2;
bound3d.maxx:=bound3d.maxx+ZAreaElementDimension_x*2;
bound3d.minz:=bound3d.minz-ZAreaElementDimension_z*2;
bound3d.maxz:=bound3d.maxz+ZAreaElementDimension_z*2;
end;

procedure DepthCollisionUseBounds(bound3d:TBounds);
begin

  min_y:=Bound3d.miny;
  max_y:=Bound3d.maxy;

 ZAreaOffset:=GetCenterOffset(bound3d);

 ZAreaDimension_x:=bound3d.maxx-bound3d.minx+ZAreaElementDimension_x;
 ZAreaDimension_z:=bound3d.maxz-bound3d.minz+ZAreaElementDimension_z;

 ZLayer_w:= ZBuf_w * ceil(ZAreaDimension_x / ZAreaElementDimension_x);
 ZLayer_h:= ZBuf_h * ceil(ZAreaDimension_z / ZAreaElementDimension_z);

 ZArea_x:= ZLayer_w div ZBuf_w;
 ZArea_y:= 16;
 ZArea_z:= ZLayer_h div ZBuf_h;

 ZBufElementDimension_x:=ZAreaElementDimension_x/ZBuf_w;
 ZBufElementDimension_z:=ZAreaElementDimension_z/ZBuf_h;

 setlength(ZArea,
           ZArea_x,
           ZArea_y,
           ZArea_z);

end;

procedure BeforeRender;
begin

 if depthcollisionFirstRun=true then
  begin
   depthcollisionFirstRun:=false;
   z_i:=0;
   z_j:=0;
   z_k:=0;
   SetOrthoMode;
  end;

 glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
 glLoadIdentity();

 SetOrthoCamera(z_i,z_j,z_k);

end;

function AfterRender:boolean;
var AllDone:boolean;
begin
 AllDone:=false;

 GetCurrentZBuffer;
 StoreBufferData(z_i,z_j,z_k);

 inc(z_i);
 if z_i>ZArea_x-1 then
  begin
   z_i:=0;
   inc(z_j);
   if z_j>ZArea_y-1 then
    begin
     z_j:=0;
     inc(z_k);
     if z_k>ZArea_z-1 then
      begin
       AllDone:=true;
       depthcollisionFirstRun:=true;
      end;
    end;
  end;

  result:=AllDone;
end;

procedure SetOrthoMode;
begin
 glMatrixMode(GL_PROJECTION);
 glPushMatrix();
 glLoadIdentity();

 glOrtho(-ZAreaElementDimension_x/2, ZAreaElementDimension_x/2,
         -ZAreaElementDimension_z/2, ZAreaElementDimension_z/2,
             my_NEAR, my_FAR);

 glMatrixMode(GL_MODELVIEW);
 glPushMatrix();
 glLoadIdentity();

 glPopMatrix;
 glPopMatrix;

 glViewport(0, 0, zbuf_W,zbuf_H);
end;

procedure SetOrthoCamera(Area_i,Area_j,Area_k:integer);
var tmpX,tmpY,tmpZ:single;
begin

  tmpY:=min_Y + ( Area_j/(ZArea_y-1) ) * (max_y - min_y);

  tmpX:=Area_i*ZAreaElementDimension_x - 0.5*ZAreaDimension_x + 0.5*ZBufElementDimension_x + ZAreaOffset.x;
  tmpZ:=Area_k*ZAreaElementDimension_z - 0.5*ZAreaDimension_z + 0.5*ZBufElementDimension_z + ZAreaOffset.z;

//tmpX:=(Area_i - (ZArea_x-1)/2)*ZAreaElementDimension_x;
//tmpZ:=(Area_j - (ZArea_z-1)/2)*ZAreaElementDimension_z;



  gluLookAt(tmpX, tmpY  , tmpZ,
            tmpX, tmpY-1, tmpZ,
            1,    0,      0);
end;

procedure GetCurrentZBuffer;
begin
  glReadPixels(0, 0, zbuf_W, zbuf_H, GL_DEPTH_COMPONENT, GL_FLOAT, @Zbuffer);
end;

procedure StoreBufferData(Area_i,Area_j,Area_k:integer);
var k:integer;
    tmpY:single;

    i,j:integer;
    FOUNDEQUAL:boolean;
    EQUALNUMBER:integer;
    EQUALDIFF:single;
    ANOTHERHEIGHT:single;
begin

 FOUNDEQUAL:=FALSE;

 
 for i:=0 to length(ZLayerElements)-1 do
  if isequal(ZBuffer,ZLayerElements[i].zDATA) then
   begin
    FOUNDEQUAL:=true;
    EQUALNUMBER:=i;
    EQUALDIFF:=getDIFF(ZBuffer,ZLayerElements[i].zDATA);
    ANOTHERHEIGHT:=ZLayerElements[i].captureHeight;
   end;

 if not FOUNDEQUAL then
  begin
   k:=length(ZLayerElements);
   setlength(ZLayerElements,k+1);

   tmpY:=min_Y + ( Area_j/(ZArea_y-1) ) * (max_y - min_y);

   ZLayerElements[k].zDATA:=ZBuffer;
   ZLayerElements[k].captureHeight:=tmpY;
   ZLayerElements[k].parent_i:=Area_i;
   ZLayerElements[k].parent_j:=Area_j;
   ZLayerElements[k].parent_k:=Area_k;

   ZArea[Area_i,Area_j,Area_k].child_index:=k;
   ZArea[Area_i,Area_j,Area_k].captureHeight:=tmpY;
   // ZArea[Area_i,Area_j,Area_k].optionalConst:=-1;
 end
  else
 begin
  ZArea[Area_i,Area_j,Area_k].child_index:=EQUALNUMBER;
  ZArea[Area_i,Area_j,Area_k].captureHeight:=ANOTHERHEIGHT;
 end;

end;

procedure SaveData(fname:string;bound3d:TBounds);
var f:file;
var i,j,k:integer;
begin
 assignFile(f,fname);
 rewrite(f,1);


 blockWrite(f,bound3d,sizeof(bound3d));

 i:=length(ZLayerElements);blockWrite(f,i,sizeof(i));


 for i:=0 to ZArea_x-1 do
  for j:=0 to ZArea_y-1 do
   for k:=0 to ZArea_z-1 do
    blockWrite(f,ZArea[i,j,k],SizeOf(TLinkToZLayerElement));
//  blockWrite(f,ZArea,SizeOf(TLinkToZLayerElement)*ZArea_x*ZArea_y*ZArea_Z);
   for i:=0 to length(ZLayerElements)-1 do
   begin
    blockWrite(f,ZLayerElements[i],sizeof(TZLayerElement))
   end;


 closeFile(f);
end;


procedure LoadData(fname:string;var bound3d:TBounds);
var f:file;
var i,j,k:integer;
begin
 assignFile(f,fname);
 reset(f,1);


 blockRead(f,bound3d,sizeof(bound3d));
 DepthCollisionUseBounds(bound3d);
 blockRead(f,i,sizeof(i));setlength(ZLayerElements,i);

 for i:=0 to ZArea_x-1 do
  for j:=0 to ZArea_y-1 do
   for k:=0 to ZArea_z-1 do
    blockRead(f,ZArea[i,j,k],SizeOf(TLinkToZLayerElement));

   for i:=0 to length(ZLayerElements)-1 do
   begin
    blockRead(f,ZLayerElements[i],sizeof(TZLayerElement))
   end;


 closeFile(f);
end;


procedure ShowCollisionDataOpengl(ZoneI,ZoneJ,ZoneK:integer; ZoneU,ZoneV:integer; Player_x,Player_y,Player_z:single);
var Area_i,Area_j,Area_k:integer;
var tmpX,tmpY,tmpZ:single;
var u,v:integer;
begin



 for Area_i:=0 to ZArea_x-1 do
  for Area_j:=0 to ZArea_y-1 do
   for Area_k:=0 to ZArea_z-1 do
    begin

 if not ( (area_i=zonei)and(area_j=zonej)and(area_k=zonek) ) then continue;

    GetXYZfromIJK(tmpX,tmpY,tmpZ,
                  Area_i,Area_j,Area_k);

     tmpY:=tmpY + ZAreaOffset.y;         

     glColor3f(1,1,1);
     glBegin(GL_LINES);
      glVertex3f(tmpX,tmpY,tmpZ);
      glVertex3f(tmpX,tmpY-10,tmpZ);
     glEnd;

     glPointSize(3);
     if ( (area_i=zonei)and(area_j=zonej)and(area_k=zonek) ) then
      glColor3f(0,0,1) else glColor3f(1,0,0);

     glBegin(GL_POINTS);
      glVertex3f(tmpX,tmpY,tmpZ);
      glVertex3f(Player_x,Player_y,Player_z);
     glEnd;



     if ( (area_i=zonei)and(area_j=zonej)and(area_k=zonek) ) then
     glColor3f(0,1,0) else glColor3f(1,0,1);
(*
     glColor3f(Area_j/ZArea_y,
               Area_j/ZArea_y,
               Area_j/ZArea_y);
*)

     for u:=0 to ZBuf_w-1 do
      for v:=0 to ZBuf_h-1 do
       begin

       GetXYZfromIJKUV(tmpX,tmpY,tmpZ,
                       Area_i,Area_j,Area_k,
                       u,v);


        if ( (area_i=zonei)and(area_j=zonej)and(area_k=zonek)and(u=zoneU)and(v=zoneV) ) then
        glPointSize(5) else   glPointSize(1);
        glBegin(GL_POINTS);
        glVertex3f(tmpX,tmpY,tmpZ);
        glEnd;

        if ( (area_i=zonei)and(area_j=zonej)and(area_k=zonek)and(u=zoneU)and(v=zoneV) ) then
         begin
          glBegin(GL_LINES);
           glVertex3f(Player_x,Player_y,Player_z);
           glVertex3f(tmpX,tmpY,tmpZ);
          glEnd;
         end;


       end;

    end;

    glColor3f(1,1,1);
end;

function isequal(zbuf1,zbuf2:TZbuffer):boolean;
var k,l:integer;
    diff:single;
    ISequal:boolean;
begin
 diff:=zbuf2[0,0]-zbuf1[0,0];
 ISequal:=true;

 for k:=0 to ZBuf_w-1 do
  for l:=0 to ZBuf_h-1 do
   begin
    if abs(zbuf2[k,l]-zbuf1[k,l])>EPSILON then ISequal:=false;
   end;

 result:=ISequal;   
end;

function getDIFF(zbuf1,zbuf2:TZbuffer):single;
begin
 result:=zbuf2[0,0]-zbuf1[0,0];
end;


end.
