// -------------------------------------------------------------
// MYENGINE3D v2008.2.20 by Georgy Moshkin
// licence: public domain
// -------------------------------------------------------------
unit myquat;

interface

uses myvectors,mymatrix;

type TQuat=record
            x,y,z,w:single;
           end;

function QuatFromAngles(angles:TVector):TQuat;
function QuatLerp(q1, q2: TQuat; t: single): TQuat;
function QuatMat(q: TQuat):TMatrix;

implementation

function QuatFromAngles(angles:TVector):TQuat; // from hlsmd?

var angle             : single;
    sr,sp,sy,cr,cp,cy : single;
    crcp,srsp         : single;

begin
     angle:=angles.z*0.5;
     sy:=sin( angle );
     cy:=cos( angle );
     angle:= angles.y*0.5;
     sp:= sin( angle );
     cp:= cos( angle );
     angle:= angles.x*0.5;
     sr:= sin( angle );
     cr:= cos( angle );

     crcp:= cr*cp;
     srsp:= sr*sp;

     result.x:=sr*cp*cy-cr*sp*sy;
     result.y:=cr*sp*cy+sr*cp*sy;
     result.z:=crcp*sy-srsp*cy;
     result.w:=crcp*cy+srsp*sy;
end;

function QuatLerp(q1, q2: TQuat; t: single): TQuat; // based on xproger's source
var dot,DOTxT:single;
begin

 dot:=q1.x*q2.x+q1.y*q2.y+q1.z*q2.z+q1.w*q2.w;

 if dot<0 then dot:=-1 else dot:=+1;
 DOTxT:=dot*t;

 result.x:=q1.x+DOTxT*(q2.x-dot*q1.x);
 result.y:=q1.y+DOTxT*(q2.y-dot*q1.y);
 result.z:=q1.z+DOTxT*(q2.z-dot*q1.z);
 result.w:=q1.w+DOTxT*(q2.w-dot*q1.w);

end;

function QuatMat(q: TQuat):TMatrix; // from delphi3d.net
var
  wx, wy, wz, xx, yy, yz, xy, xz, zz, x2, y2, z2 : single;
  m:TMatrix;
begin
  with q do
  begin
    x2 := x + x;   y2 := y + y;   z2 := z + z;
    xx := x * x2;  xy := x * y2;  xz := x * z2;
    yy := y * y2;  yz := y * z2;  zz := z * z2;
    wx := w * x2;  wy := w * y2;  wz := w * z2;
  end;

  m[0,0] := 1 - (yy + zz);  m[1,0] := xy + wz;        m[2,0] := xz - wy;
  m[0,1] := xy - wz;        m[1,1] := 1 - (xx + zz);  m[2,1] := yz + wx;
  m[0,2] := xz + wy;        m[1,2] := yz - wx;        m[2,2] := 1 - (xx + yy);

  m[3,0] := 0;
  m[3,1] := 0;
  m[3,2] := 0;
  m[0,3] := 0;
  m[1,3] := 0;
  m[2,3] := 0;
  m[3,3] := 1;

  result:=m;


end;




end.
